package main;
import java.util.*;
import java.io.*;
import java.awt.Color;
import java.text.DecimalFormat;

/**
 * Class for presenting GraphNodes as Digraphs using dot
 */
public class Digraph {

    public enum DigraphMode {
	NORMAL, COLLAPSE, COLLAPSEMULTIEDGE, TREESTRETCH, UNDIRECTEDCOLLAPSE;
    }

    DigraphMode mode;

    //boolean uniquepaths;
    //boolean uniqueedges;
    boolean flipedges;
    DepthStretchSettings ds;

    // Set<String> nodeset = null;
    //Set<TwoTuple<String,String>> edgeset = null;
    
    public Digraph(DigraphMode mode, boolean flip, DepthStretchSettings d) {
	//uniquepaths = up;
	//uniqueedges = ue;
	
	switch(mode) {
	case NORMAL:
	case COLLAPSE:
	case COLLAPSEMULTIEDGE:
	case UNDIRECTEDCOLLAPSE:
	    if(d != null) {
		throw new RuntimeException("Non-null DepthStretch for mode "+mode);
	    }
	    break;
	case TREESTRETCH:
	    if(d == null) {
		throw new RuntimeException("Null DepthStretch for mode "+mode);
	    }
	    break;
	default:
	    throw new RuntimeException("Unknown mode "+mode);
	}
	ds        = d;
	this.mode = mode;
	flipedges = flip;
	
    }

    public Digraph(DigraphMode mode, DepthStretchSettings d) {
	this(mode, false, d);
    }

    public Digraph(DigraphMode mode, boolean flip) {
	this(mode, flip, null);
    }

    public Digraph(DigraphMode mode) {
	this(mode, false);
    }

    public Digraph(DepthStretchSettings d) {
	this(DigraphMode.TREESTRETCH, false, d);
    }

    public Digraph() {
	this(DigraphMode.NORMAL);
    }

//     public Digraph(DepthStretchSettings hs) {
// 	this(true,false,false,hs);
//     }
    
//     public Digraph() {
// 	this(null);
//     }



    public <E> String  getDigraph(String name, GraphNode<E> graph) {
	StringBuffer b = new StringBuffer();
	DigraphWalker gw = new DigraphWalker(b);

	GraphWalkerWalkDetails wd;
	switch(mode) {
	case NORMAL:
	case COLLAPSE:
	case COLLAPSEMULTIEDGE:
	case UNDIRECTEDCOLLAPSE:
	    wd = new GraphWalkerWalkDetails(true,true,null);
	    break;
	case TREESTRETCH:
	    wd = new GraphWalkerWalkDetails(true,false,null);
	    break;
	default:
	    throw new RuntimeException("Unknown mode "+mode);
	}
	gw.walk(graph,wd);
	return "digraph "+name+" {\n"+b.toString()+"}\n";
    }

    public <E> String  getDigraph(GraphNode<E> graph) {
	return getDigraph("G",graph);
    }


    /* ============================================================ */

    class DigraphWalker extends GraphWalker {

	StringBuffer b;
	Set<String> nodeset = new HashSet<String>();
	Set<TwoTuple<String,String>> edgeset = new HashSet<TwoTuple<String,String>>();


	public DigraphWalker(StringBuffer b) {
	    this.b = b;
	}

	public <E> String getNodeName(GraphNode<E> node) {
	    switch(mode) {
	    case NORMAL:
	    case TREESTRETCH:
		return "n"+node.getNodeId();
	    case COLLAPSE:
	    case COLLAPSEMULTIEDGE:
	    case UNDIRECTEDCOLLAPSE:
		return "c"+node.contentToString();
	    default:
		throw new RuntimeException("Unknown mode "+mode);
	    }
	}

	public <E> String getFakeNodeName(GraphNode<E> from, GraphNode<E> to, int number) {
	    switch(mode) {
	    case TREESTRETCH:
		return "f"+from.getNodeId()+"_"+to.getNodeId()+"_"+number;
	    default:
		throw new RuntimeException("Fake nodes unsupported in mode "+mode);
	    }
	}
	

	public <E> void putNode(GraphNode<E> node) {
	    b.append("   "+getNodeName(node)+" [label=\""+node.contentToString()+"\"]\n");
	}
    

	public <E> void putFakeNode(String fakename) {
	    b.append("   "+fakename+" [label=\"\", width=0.01, height=0.01]\n");
	}

	/**
	 * Put a stretched edge with <code>amount</code> extra fake
	 * nodes on the path from <code>from</code> to
	 * <code>to</code>.
	 */
	public <E> void putStretchedEdge(GraphNode<E> from, GraphNode<E> to, int amount) {
	    String fromname = getNodeName(from);
	    if(amount > 0) {
		for(int i=0; i < amount; i++) {
		    String fakename = getFakeNodeName(from,to,i);
		    putFakeNode(fakename);
		    boolean arrow = (i == 0 && flipedges) || (i == amount -1 && !flipedges);
		    putEdge(fromname,fakename,arrow);
		    fromname = fakename;
		}
	    }
	    putEdge(fromname,getNodeName(to));
	}

	public  <E> void putEdge(String from, String to, boolean arrow) {
	    String options;
	    if(arrow) {
		options = (flipedges ? " [dir=\"back\"]" : "");
	    } else {
		options = " [arrowhead=none]";
	    }
	    b.append("   "+from+" -> "+to+options+"\n");
	}

	public  <E> void putEdge(String from, String to) {
	    putEdge(from, to, true);
	}

	public  <E> void putEdge(GraphNode<E> from, GraphNode<E> to, boolean arrow) {
	    putEdge(getNodeName(from),getNodeName(to),arrow);
	}

	public  <E> void putEdge(GraphNode<E> from, GraphNode<E> to) {
	    putEdge(getNodeName(from),getNodeName(to),true);
	}


	public <E> boolean shallPutCollapseEdge(GraphNode<E> from, GraphNode<E> to, boolean directed) {
	    TwoTuple<String,String> edge = new TwoTuple<String,String>(getNodeName(from),getNodeName(to));
	    if(!edgeset.contains(edge)) {
		if(directed) {
		    edgeset.add(edge);
		    return true;
		} else {
		    TwoTuple<String,String> reverseedge = new TwoTuple<String,String>(getNodeName(to),getNodeName(from));
		    if(!edgeset.contains(reverseedge)) {
			edgeset.add(edge);
			return true;
		    }
		}
	    }
	    return false;
	}

	public <E> void visitNode(GraphNode<E> parent, GraphNode<E> node, GraphWalkerNodeDetails details, GraphWalkerWalkDetails walkdetails) {
	    if(details.nstate == GraphWalker.NodeState.START) {
		if(!details.revisit) {
		    boolean put = true;
		    switch(mode) {
		    case NORMAL:
		    case TREESTRETCH:
			break;
		    case COLLAPSE:
		    case COLLAPSEMULTIEDGE:
		    case UNDIRECTEDCOLLAPSE:
			String name = getNodeName(node);
			if(!nodeset.contains(name)) {
			    nodeset.add(name);
			} else {
			    put = false;
			}
			break;
		    default:
			throw new RuntimeException("Unknown mode "+mode);
		    }
		    if(put) {
			putNode(node);
		    }
		}
		if(!details.usedlink && parent != null) {
		    GraphNode<E> from, to;
		    switch(details.pstate) {
		    case FORWARD:
			from = parent;
			to   = node;
			break;
		    case BACKWARD:
			from = node;
			to   = parent;
			break;
		    default:
			throw new RuntimeException("Unknown ParentState "+details.pstate);
		    }
		    
		    boolean put = true;
		    switch(mode) {
		    case NORMAL:
		    case COLLAPSEMULTIEDGE:
			break;
		    case TREESTRETCH:
			switch(details.pstate) {
			case FORWARD:
			    putStretchedEdge(from,to,ds.getChildStretch(details.depth - 1, details.relativenumber));
			    put = false;
			    break;
			case BACKWARD:
			    throw new RuntimeException("Mode "+mode+" should never walk backwards");
			default:
			    throw new RuntimeException("Unknown ParentState "+details.pstate);
			}
			break;
		    case COLLAPSE:
			put = shallPutCollapseEdge(from,to,true);
			break;
		    case UNDIRECTEDCOLLAPSE:
			put = shallPutCollapseEdge(from,to,false);
			break;
		    default:
			throw new RuntimeException("Unknown mode "+mode);
		    }

		    if(put) {
			putEdge(from, to, !mode.equals(DigraphMode.UNDIRECTEDCOLLAPSE));
		    }
		}    
	    }
	}
    }

    /* ============================================================ */


     public static <E> void writeGraphToImage(Digraph dg, GraphNode<E> graph, String filebase,String filetype, boolean spring) {
	String graphstring = dg.getDigraph(graph);
	graphstringToImage(graphstring, filebase, filetype, spring);
     }

    public static <E> void popupDigraph(Digraph dg, GraphNode<E> graph, String title, boolean spring) {
	String graphstring = dg.getDigraph(graph);
	popupGraphstring(graphstring, spring, title);
    }

    public static <E> void popupDigraph(Digraph dg, GraphNode<E> graph, String title) {
	popupDigraph(dg,graph,title,false);
    }

     public static <E> void popupDigraph(GraphNode<E> graph, String title,boolean spring) {
 	popupDigraph(new Digraph(),graph,title,spring);
     }

     public static <E> void popupDigraph(GraphNode<E> graph, String title) {
 	popupDigraph(graph,title,false);
     }

    /* ============================================================ */
    /* Generic functions */
    /* ============================================================ */
    
    public static void stringToFile(String content, String filename) {
	try {
	    PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter(filename)));
	    out.print(content);
	    out.close();
	} catch(Exception e) {
	    e.printStackTrace();
	    System.exit(1);
	}
    }
    

    public static String dotfileToImage(String namebase, String filetype, boolean useneato) {
	String imagefile = "";
	try {
	    String dotfile   = namebase+".dot";
	    imagefile = namebase+"."+filetype;
	    String program = (useneato ? "neato" : "dot");
	    String cmd = program+" -T "+filetype+" -o "+imagefile+" "+dotfile;
	    Runtime run = Runtime.getRuntime() ;
	    Process pr = run.exec(cmd);
	    pr.waitFor();
	} catch(Exception e) {
	    e.printStackTrace();
	    System.exit(1);
	}
	return imagefile;
    }

    public static String graphstringToImage(String graphstring, String filebase, String filetype, boolean useneato) {
	stringToFile(graphstring, filebase+".dot");
	return dotfileToImage(filebase, filetype, useneato);
    }
    
    public static void popupGraphstring(String graphstring, boolean useneato, String title) {
	try {
	    String filebase  = "Digraph.popupimage";
	    String filetype  = "png";
	    String imagefile = graphstringToImage(graphstring, filebase, filetype, useneato);
	    new ImagePopup().popupImage(imagefile,title);
	} catch(Exception e) {
	    e.printStackTrace();
	    System.exit(1);
	}
    }


    /**
     * Format a float to a to decimal string
     */
    private static String twoDecimals(float f) {
	DecimalFormat df = new DecimalFormat("0.00");
	return df.format(f);
    }

    public static <E> String getColorString(Map<E,Color> colormap, E id) {
	String colorstring="";
	if(colormap != null && colormap.containsKey(id)) {
	    Color col = colormap.get(id);
	    int cr = col.getRed();
	    int cg = col.getGreen();
	    int cb = col.getBlue();
	    float[] hsb = Color.RGBtoHSB(cr, cg, cb, null);
	    String ccr = twoDecimals(hsb[0]);
	    String ccg = twoDecimals(hsb[1]);
	    String ccb = twoDecimals(hsb[2]);
 	    colorstring = ", style=filled, color=\""+ccr+" "+ccg+" "+ccb+"\"";
	}
	return colorstring;
    }
}





//     public String toDigraph(HeightStretchSettings hs) {
// 	return toDigraph(true,hs);
//     }

//     public String toDigraph() {
// 	// HeightStretchSettings hs = new HeightStretchSettings(0,0,1,4,0,8);
// 	HeightStretchSettings hs = new HeightStretchSettings(1);
// 	return toDigraph(hs);
//     }





   
//     private String pathAdd(String parent,String node) {
// 	return (parent == null ? "" : parent)+"_"+node;
//     }

//     private String getId(String parent,String node) {
// 	return (uniquepaths ? pathAdd(parent,node) : node);
//     }

//     private String newParent(String parent,String node) {
// 	return getId(parent,node);
//    }

//     private getChildId(String path,String parent, String child) {
// 	return (uniquepaths ? pathAdd(pathAdd(path,parent),child) : child);
//     }
	


//     /**
//      * Outputs a node line. If label is <code>null</code> a "fake" node will be put.
//      */
//     private String putDigraphNode(String parent,String node,String label) {
// 	String id = getId(parent,node);
// 	String options = (label != null ? "label=\""+label+"\"" : "label=\"\", width=0.01, height=0.01");
// 	return "\t"+id+" ["+options+"];\n";
//     }
    
//     private String putDigraphEdge(String parent, String node) {
// 	if(parent != null && !parent.equals("")) {
// 	    boolean add = true;
// 	    if(uniqueedges) {
// 		TwoTuple<String,String> edge = new TwoTuple<String,String>(parent,node);
// 		if(edgeset.contains(edge)) {
// 		    add = false;
// 		} else {
// 		    edgeset.add(edge);
// 		}
// 	    }
// 	    if(add) {
// 		String from = parent; 
// 		String to = getId(parent,node);
// 		String options = (flipedges ? " [dir=\"back\"]" : "");
// 		return "\t"+from+" -> "+to+options+";\n";
// 	    }
// 	}
// 	return "";
//     }

//     private String putFakeEdge(String parentpath,String nodename) {
// 	return "\t"+parentpath+" -> "+pathAdd(parentpath,nodename)+" [arrowhead=none];\n";
//     }
    


//     /**
//      * Add (to <code>b</code>) <code>amount</code> extra fake nodes on the path from
//      * <code>nodename</code> to <code>childnode</code>, where
//      * <code>parentpath</code> is the path to but not including
//      * nodename.
//      *
//      * @returns the new path going through those.
//      */
//     public String stretchEdge(StringBuffer b,String path,String node,String childnode,int amount) {
// 	String newparent = newParent(path,node);
// 	if(amount > 0) {
// 	    for(int i=0; i < amount; i++) {
// 		String fakenode = "fake"+node+"fake"+childnode+"fake";
// 		b.append(putDigraphNode(newparent,fakenode,null));
// 		b.append(putFakeEdge(newparent,fakenode));
// 		newparent = newParent(newparent,fakenode);
// 	    }
// 	}
// 	return newparent;
	
//     }
    
//     public <E> void fillToDigraph(StringBuffer b,String parent, TreeNode<E> tree) {
// 	String nodename = tree.contentToString();
// 	// b.append("/* Got fillToDigraph("+parent+","+tree+") */\n");
// 	b.append(putDigraphNode(parent,nodename,nodename));
// 	b.append(putDigraphEdge(parent,nodename));
// 	int childnumber = 0;
// 	for(TreeNode<E> child : tree.getTreeChildren()) {
// 	    String newparent = newParent(parent,nodename);
// 	    if(hs.stretch) {
// 		int amount = hs.getChildStretch(tree.height(),childnumber);
// 		newparent = stretchEdge(b, parent, nodename, ""+child.content, amount);
// 		childnumber++;
// 	    }
// 	    fillToDigraph(b, newparent, child);
// 	}
//     }
    
//     public <E> String getDigraph(TreeNode<E> tree) {
// 	StringBuffer b = new StringBuffer();
// 	if(uniqueedges) {
// 	    edgeset = new HashSet<TwoTuple<String,String>>();
// 	}
// 	b.append("digraph G {\n");
// 	fillToDigraph(b,null,tree);
// 	b.append("}\n");
// 	return b.toString();
//     }
