package main;
import java.util.*;

/**
 * Note that if you are walking with one GraphWalker, it is not safe
 * to invoke another one. For instance, GraphNode.toString() while
 * inside a walker helper will screw things up.
 */
public class GraphWalker {

    public enum ParentState {
	FORWARD, BACKWARD;
	
	ParentState flip() {
	    switch(this) {
	    case FORWARD:
		return BACKWARD;
	    case BACKWARD:
		return FORWARD;
	    default:
		throw new RuntimeException("Unknown ParentState "+this);
	    }
	}
    };

    public enum NodeState   {START, BEFORECHILDREN, AFTERCHILDREN, BEFOREPARENTS, AFTERPARENTS, END};

    Set<Integer> visited;

    public <E> void walk(GraphNode<E> node, GraphWalkerWalkDetails walkdetails) {
	visited = new HashSet<Integer>(); /* TODO: Better to have an array of booleans with size of the largest issued id and a way to reset that id */
	GraphWalkerNodeDetails details = new GraphWalkerNodeDetails();
	details.firstnode = true;
	details.pstate    = GraphWalker.ParentState.FORWARD;
	details.depth     = 0;
	walk(null,node,details,walkdetails);
    }
    
    protected <E> void walk(GraphNode<E> from, GraphNode<E> node, GraphWalkerNodeDetails details, GraphWalkerWalkDetails walkdetails) {
	visited.add(node.getNodeId());
	
	if(!details.revisit) {
	    node.wh.resetLinks();
	}
	if(from != null) {
	    node.wh.useLink(from,details.pstate.flip());
	}
	
	details.nstate = GraphWalker.NodeState.START;
	visitNode(from, node, details, walkdetails);

	if(!details.revisit) {
	    if(walkdetails.forward) {
		details.nstate = GraphWalker.NodeState.BEFORECHILDREN;
		visitNode(from, node, details, walkdetails);
	    
		walkList(node, node.getChildren(), GraphWalker.ParentState.FORWARD, details, walkdetails);
	    
		details.nstate = GraphWalker.NodeState.AFTERCHILDREN;
		visitNode(from, node, details, walkdetails);
	    }
	
	    if(walkdetails.backward) {
		details.nstate = GraphWalker.NodeState.BEFOREPARENTS;
		visitNode(from, node, details, walkdetails);
	    
		walkList(node, node.getParents(), GraphWalker.ParentState.BACKWARD, details, walkdetails);
	    
		details.nstate = GraphWalker.NodeState.AFTERPARENTS;
		visitNode(from, node, details, walkdetails);
	    }
	}

	details.nstate = GraphWalker.NodeState.END;
	visitNode(from, node, details, walkdetails);

    }




//     protected boolean isUsedLink(GraphNode<E> node, GraphNode<E> n, GraphWalker.ParentState pstate) {
// 	switch(pstate) {
// 	case FORWARD:
// 	    return node.wh.isUsedLink(n);
// 	case BACKWARD:
// 	    return n.wh.isUsedLink(node);
// 	default:
// 	    throw new RuntimeException("Unknown ParentState");
// 	}
//     }

//     protected void usedLink(GraphNode<E> node, GraphNode<E> n, GraphWalker.ParentState pstate) {
// 	switch(pstate) {
// 	case FORWARD:
// 	    return node.wh.useLink(n);
// 	case BACKWARD:
// 	    return n.wh.useLink(node);
// 	default:
// 	    throw new RuntimeException("Unknown ParentState");
// 	}
//     }
    
    protected <E> void walkList(GraphNode<E> node, List<GraphNode<E>> list, GraphWalker.ParentState pstate, GraphWalkerNodeDetails details, GraphWalkerWalkDetails walkdetails) {
	for(int i=0; i < list.size(); i++) {
	    GraphNode<E> n = list.get(i);
	    GraphWalkerNodeDetails d = new GraphWalkerNodeDetails();
	    d.firstrelative  = (i == 0);
	    d.lastrelative   = (i == list.size() - 1);
	    d.relativenumber = i;
	    d.revisit        = visited.contains(n.getNodeId());
	    d.pstate         = pstate;
	    d.depth          = details.depth + 1;
	    d.usedlink       = node.wh.isUsedLink(n,pstate);
	    walk(node, n, d, walkdetails);
	}
    }

    public <E> void visitNode(GraphNode<E> parent, GraphNode<E> node, GraphWalkerNodeDetails details, GraphWalkerWalkDetails walkdetails) {
	throw new RuntimeException("visitNode is to be overridden!");
    }
    





    private static final int indentationstep = 5;
    private static int debugindentation = 0;
    private static void indentInc() {
	debugindentation += indentationstep;
    }
    private static void indentDec() {
	debugindentation -= indentationstep;
    }
    private static void indent(StringBuffer b) {
	for(int i=0; i < debugindentation; i++) {
	    b.append(" ");
	}
    }

    public static <E> String debugWalk(GraphNode<E> graph, boolean forward, boolean backward) {
	StringBuffer b = new StringBuffer();
	GraphWalker gw = new GraphWalker() 
	    {
		public <E> void visitNode(GraphNode<E> parent, GraphNode<E> node, GraphWalkerNodeDetails details, GraphWalkerWalkDetails walkdetails) {
		    Object o = walkdetails.visitobject;
		    if(details.firstrelative && details.nstate == GraphWalker.NodeState.START) {
			indentInc();
		    }
		    if(o instanceof StringBuffer) {
			StringBuffer b = (StringBuffer)o;
			String pstring = (parent == null ? "null" : parent.contentToString());
			indent(b);
			b.append("visitNode("+pstring+","+node.contentToString()+",...)");
			b.append(" - "+details+"\n");
		    }
		    if(details.lastrelative && details.nstate == GraphWalker.NodeState.END) {
			indentDec();
		    }
		}
	    };
	GraphWalkerWalkDetails wd = new GraphWalkerWalkDetails(forward,backward,b);
	gw.walk(graph, wd);
	return b.toString();
    }

}