package main;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Map;
import java.util.Random;
import java.util.SortedMap;
import java.util.TreeMap;

public class Network {

    SortedMap<Coordinate,NetworkNode> nodes;

    public long netseed;
    Random netrand;

    public long simseed;
    Random simrand;

    int minx;
    int maxx;
    int miny;
    int maxy;
    
    public long arbseed;
    public boolean arbitrarystate;
    Random arbrand;

    public boolean badstate;
    public int badstateinput;
    
    public Network(long ns, long ss) {
	nodes  = new TreeMap<Coordinate,NetworkNode>();
	netseed = ns;
	netrand = new Random(netseed);
	setSimSeed(ss);
        arbitrarystate = false;
        arbrand = null;
    }


    public Network(long ns, long ss, long as) {
        this(ns, ss);
        setArbSeed(as);
    }

    public Network() {
	this(new Random().nextLong(), new Random().nextLong());
    }
    
    public Network(SortedMap<Coordinate,NetworkNode> nodes, long ss) {
	this(0, ss);
	this.nodes = nodes;
	for(NetworkNode node : nodes.values()) {
	    node.setNetwork(this);
	}
	
    }

    public Network(SortedMap<Coordinate,NetworkNode> nodes) {
	this(nodes, new Random().nextLong());
    }
    
    public void setSimSeed(long ss) {
        simseed = ss;
	simrand = new Random(simseed);
    }

    public void setArbSeed(long as) {
        arbseed = as;
	arbrand = new Random(arbseed);
    }

    private void ensureArbRand() {
        if(arbrand == null) {
            setArbSeed(new Random().nextLong());
        }
    }
    
    public void setArbitraryState() {
        arbitrarystate = true;
        ensureArbRand();
        for(NetworkNode n : nodes.values()) {
            n.setArbitraryState(arbrand);
        }
    }

    public void setBadState(int input) {
        badstate = true;
        badstateinput = input;
        ensureArbRand();
        for(NetworkNode n : nodes.values()) {
            n.setBadState(arbrand, input);
        }
    }


    /* ============================================================ */


//     public void addNode(Coordinate coord,NetworkNode node) {
// 	nodes.put(coord, node);
//     }

    public final SortedMap<Coordinate,NetworkNode> getNodes() {
	return nodes;
    }

    public void stepNetwork() {
	for(NetworkNode n : nodes.values()) {
	    n.stepNode();
	}
	for(NetworkNode n : nodes.values()) {
	    ArrayList<NetworkMessage> inbox = n.getInbox();
	    while(inbox.size() > 0) {
		int index = simrand.nextInt(inbox.size());
		NetworkMessage m = inbox.remove(index);
		n.receiveMessage(m);
	    }
	}
    }
    
    /**
     * Node <code>node</code> broadcasts message <code>mess</code>
     */
    public void broadcast(NetworkNode node,NetworkMessage mess) {
	for(NetworkNode n : node.getNeighbors()) {
	    n.storeMessage(mess.getCopy());
	}

    }
    


    public GraphNode<Integer> getNetworkGraph() {
	GraphNode<Integer> first = null;
	for(NetworkNode n : nodes.values()) {
	    n.tmpgn = new GraphNode<Integer>(n.getNumber());
	    if(first == null) {
		first = n.tmpgn;
	    }
	}
	for(NetworkNode n : nodes.values()) {
	    for(NetworkNode o : n.getNeighbors()) {
		n.tmpgn.addChild(o.tmpgn);
	    }
	}	
	for(NetworkNode n : nodes.values()) {
	    n.tmpgn = null;
	}
	return first;
    }
    
    public String getDigraph() {
	return new Digraph().getDigraph(getNetworkGraph());
    }

    public Collection<NetworkNode> getNetworkNodes() {
	return  nodes.values();
    }

    public NetworkNode getANode() {
        for(NetworkNode n : nodes.values()) {
            return n;
        }
        throw new RuntimeException("No nodes in network!");
    }


    /* ==================== Building Networks ==================== */

    /**
     * @return the seed used to generate the network
     */
    private void internalBuildNetwork(NetworkNodeFactory factory, int numnodes, int areaside, int range) {
        FileWriter file = null;
        PrintWriter pw = null;

            try {
				file = new FileWriter("./tos_topo.txt");
			} catch (IOException e) {
				System.out.println("Error while trying to open the file");
				e.printStackTrace();
			}
            pw = new PrintWriter(file);
    	
	if(numnodes > areaside * areaside) {
	    throw new RuntimeException("Insufficient area to put "+numnodes+" nodes in a "+areaside+" x "+areaside+" square");
	}
	int i = 0;
	while(i < numnodes) {
	    Coordinate c = new Coordinate(netrand.nextInt(areaside),netrand.nextInt(areaside));
	    if(!nodes.containsKey(c)) {
		nodes.put(c, factory.newNode(this));
		i++;
	    }
            
	}

	pw.println("nodes\t" + i);
	
	for(Coordinate c : nodes.keySet()) {
	    for(Coordinate c2 : nodes.keySet()) {
	    	if(!c.equals(c2) && c.squareDistanceTo(c2) <= range*range) {
	    		NetworkNode n = nodes.get(c);
	    		NetworkNode n2 = nodes.get(c2);
	    		n.addNeighbor(n2);
	    		pw.println("gain\t" + n.getNumber() + "\t" + n2.getNumber() + "\t" + "-20.0");
	    	}
	    }
	}
	try {
		file.close();
	} catch (IOException e) {
		System.out.println("Error while trying to close the file");
		e.printStackTrace();
	}
    }

    public static void buildNetwork(Network net, NetworkNodeFactory factory, int numnodes, int areaside, int range) {
	net.internalBuildNetwork(factory, numnodes, areaside, range);
    }

    public static Network buildNetwork(NetworkNodeFactory factory, int numnodes, int areaside, int range) {
	Network net = new Network();
	buildNetwork(net, factory, numnodes, areaside, range);
	return net;
    }

    /* ==================== */

    private static void buildNetwork(Network net, NetworkNodeFactory factory, double nodespercircle, int areaside, int range) {
	double area = areaside * areaside;
	double radiocover = (range * range) * Math.PI;
	int numnodes = (int)(area / radiocover * nodespercircle);
	net.internalBuildNetwork(factory, numnodes, areaside, range);
    }

    public static Network buildNetwork(NetworkNodeFactory factory, double nodespercircle, int areaside, int range) {
	Network net = new Network();
	buildNetwork(net, factory, nodespercircle, areaside, range);
	return net;
    }

    public static Network buildNetwork(NetworkNodeFactory factory, double nodespercircle, int areaside, int range, long netseed, long simseed) {
	Network net = new Network(netseed, simseed);
	buildNetwork(net, factory, nodespercircle, areaside, range);
	return net;
    }

    /* ================================================================================ */

    public void updateMaxMin() {
	boolean first = true;
	for(Coordinate c : nodes.keySet()) {
	    //System.out.println("c: "+c);
	    if(first) {
		minx = c.x;
		maxx = c.x;
		miny = c.y;
		maxy = c.y;
		first = false;
	    } else {		
		minx = Math.min(minx, c.x);
		maxx = Math.max(maxx, c.x);
		miny = Math.min(miny, c.y);
		maxy = Math.max(maxy, c.y);
	    }
	    //System.out.println("minx is now "+minx);
	    //System.out.println("maxx is now "+maxx);
	    //System.out.println("miny is now "+miny);
	    //System.out.println("maxy is now "+maxy);
	    //System.out.println("");
	}
    }
    
    /**
     *
     */
    public boolean hasOnlyDoubleLinks() {
	for(NetworkNode node : nodes.values()) {
            for(NetworkNode nn : node.getNeighbors()) {
                if(!nn.getNeighbors().contains(node)) {
                    System.out.println("Node "+nn+" does not have "+node+" as neighbor");
                    return false;
                }
            }
        }
        return true;
    }

    public String toString() {
	int space = 0;
	updateMaxMin();
	int nodelen = 3;
	int xspan = maxx - minx;
	int yspan = maxy - miny;
	//System.out.println("xspan "+xspan);
	//System.out.println("yspan "+yspan);
	ArrayList<ArrayList<Integer>> rows = new ArrayList<ArrayList<Integer>>();
	for(int iy = 0; iy <= yspan; iy++) {
	    rows.add(new ArrayList<Integer>());
	    for(int ix = 0; ix <= xspan; ix++) {
		rows.get(iy).add(-1);
	    }
	}
	for(Coordinate c : nodes.keySet()) {
	    ArrayList<Integer> row = rows.get(c.y-miny);
	    row.set(c.x-minx, nodes.get(c).getNumber());
	}
	StringBuffer b = new StringBuffer();
	b.append('+');
	for(int ix = 0; ix < (xspan+1) * (nodelen + space); ix++) {
	    b.append('-');
	}
	b.append("+\n");
	for(int iy = 0; iy < (yspan+1); iy++) {
	    b.append("|");
	    for(int ix = 0; ix <= xspan; ix++) {
		int num = rows.get(iy).get(ix);
		String s = "";
		if(num >= 0) {
		    s = ""+(num % ((int)Math.pow(10.0,nodelen)));
		}
		for(int i = 0; i < nodelen - s.length() + space; i++) {
		    b.append(' ');
		}
		b.append(s);
	    }
	    b.append("|\n");
	}
	b.append('+');
	for(int ix = 0; ix < (xspan + 1) * (nodelen + space); ix++) {
	    b.append('-');
	}
	b.append("+\n");
	return b.toString();
    }

    public String toJavaCodeString(String variablename) {
        StringBuffer b = new StringBuffer();
        boolean first = true;
        String buildername = "builder";
        String factoryname = "factory";
        for(Map.Entry<Coordinate,NetworkNode> entry : nodes.entrySet()) {
            Coordinate c  = entry.getKey();
            NetworkNode n = entry.getValue();
            if(first) {
                b.append(n.getFactory().toJavaCodeString(factoryname));
                b.append("NetworkBuilder "+buildername+" = new NetworkBuilder("+factoryname+");\n");
                first = false;
            }
            String neighborlist = ""+n.getNeighbors();
            b.append(buildername+".add("+n.getNumber()+","+c.x+","+c.y+", new int[] {");
            boolean firstneighbor = true;
            for(NetworkNode neighbor : n.getNeighbors()) {
                if(firstneighbor) {
                    firstneighbor = false;
                } else {
                    b.append(", ");
                }
                b.append(neighbor.getNumber());
            }
            b.append("});\n");
	}
        b.append("Network "+variablename+" = "+buildername+".getNetwork();\n");
        b.append(variablename+".setSimSeed("+simseed+"L);\n");
        if(arbitrarystate || badstate) {
            b.append(variablename+".setArbSeed("+arbseed+"L);\n");
            if(arbitrarystate) {
                b.append("net.setArbitraryState();\n");
            }
            if(badstate) {
                b.append("net.setBadState("+badstateinput+");\n");
            }
        }
        return b.toString();
    }

    public int size() {
        return nodes.size();
    }
   
}
